[hidden] {
    display: none;
}

.visually-hidden {
    position: absolute;
    clip: rect(0, 0, 0, 0);
}

div.awesomplete {
    display: inline-block;
    position: relative;
}

div.awesomplete>input {}

div.awesomplete>ul {
    position: absolute;
    z-index: 10;
    min-width: 100%;
    box-sizing: border-box;
    list-style: none;
    padding: 5px;
    margin: 0 0 0 8px;
    background: lightgray;
    opacity: .85;
    border: 1px solid rgba(0, 0, 0, .3);
    box-shadow: .05em .2em .6em rgba(0, 0, 0, .2);
    text-shadow: none;
    font-size: 10pt;
    border-bottom-left-radius: 5px;
    border-bottom-right-radius: 5px;
}

div.awesomplete>ul[hidden],
div.awesomplete>ul:empty {
    display: none;
}

@supports (transform: scale(0)) {
    div.awesomplete>ul {
        transition: .3s cubic-bezier(.4, .2, .5, 1.4);
        transform-origin: 1.43em -.43em;
    }

    div.awesomplete>ul[hidden],
    div.awesomplete>ul:empty {
        opacity: 0;
        transform: scale(0);
        display: block;
        transition-timing-function: ease;
    }
}

/* Pointer
	div.awesomplete > ul:before {
		content: "";
		position: absolute;
		top: -.43em;
		left: 1em;
		width: 0; height: 0;
		padding: .4em;
		background: white;
		border: inherit;
		border-right: 0;
		border-bottom: 0;
		-webkit-transform: rotate(45deg);
		transform: rotate(45deg);
	}*/

div.awesomplete>ul>li {
    position: relative;
    padding: .2em .5em;
    cursor: pointer;
}

div.awesomplete>ul>li:hover {
    background: hsl(200, 40%, 80%);
    color: black;
}

div.awesomplete>ul>li[aria-selected="true"] {
    background: hsl(205, 40%, 40%);
    color: white;
}

div.awesomplete mark {
    /* background: hsl(65, 100%, 50%); */
    background: cyan;
}

div.awesomplete li:hover mark {
    /* background: hsl(68, 100%, 41%); */
    background: cyan;
}

div.awesomplete li[aria-selected="true"] mark {
    background: hsl(86, 100%, 21%);
    color: inherit;
}